# -*- coding: utf-8 -*-
import warnings
from .v61240 import bscan_spec as v61240_bscan
from .v61240 import oct_spec as v61240_oct
from .v69530 import bscan_spec as v69530_bscan
from .v69530 import oct_spec as v69530_oct

from .base import bscan_base_spec, oct_base_spec


def HEXML_VERSIONS(version):
    try:
        versions = {"6.12.4.0": v61240_oct, "6.9.53.0": v69530_oct}
        return versions[version]()
    except KeyError:
        warnings.warn("The XML export version is not known.")
        return [(key, *value) for key, value in oct_base_spec().items()]


def HEXML_BSCAN_VERSIONS(version):
    try:
        versions = {"6.12.4.0": v61240_bscan, "6.9.53.0": v69530_bscan}
        return versions[version]()
    except KeyError:
        warnings.warn("The XML export version is not known.")
        return [(key, *value) for key, value in bscan_base_spec().items()]



# -*- coding: utf-8 -*-
import logging

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def compare_layers_per_ascan(data_gt, data_pred, layers=["BM", "RPE"]):
    """Compare layer annotations for two OCT volumes."""
    drusen_map = data_gt.drusen_projection > 0
    result = {}
    for layer in layers:
        dist = data_gt.layers[layer] - data_pred.layers[layer]
        for i, bscan in enumerate(data_gt):
            i += 1
            for j in range(dist.shape[1]):
                if not np.isnan(dist[-i, j]):
                    result[(f"{layer}_Distance", bscan.name, j)] = [dist[-i, j]]
                    if drusen_map[-i, j]:
                        result[(f"Drusen Load", bscan.name, j)] = "Drusen AScan"
                    else:
                        result[(f"Drusen Load", bscan.name, j)] = "No Drusen AScan"
    return result


def compare_drusen_per_bscan(data_gt, data_pred):
    """Compare drusen annotations for two OCT volumes."""
    result = {}
    for bscan_gt, bscan_pred in zip(data_gt, data_pred):
        drusen_gt = bscan_gt.drusen
        drusen_pred = bscan_pred.drusen

        union = np.sum(np.logical_or(drusen_gt, drusen_pred))
        intersection = np.sum(np.logical_and(drusen_gt, drusen_pred))
        # Square for the denominator
        denominator = np.sum(drusen_gt ** 2) + np.sum(drusen_pred ** 2)
        if denominator != 0:
            dice = (2.0 * intersection + 1e-5) / (denominator)
        else:
            dice = 1

        difference = drusen_pred.astype(int) - drusen_gt.astype(int)
        false_positives = np.sum(difference > 0)
        false_negatives = np.sum(difference < 0)
        precision = (
            intersection / (intersection + false_positives)
            if false_positives != 0
            else 1
        )
        recall = (
            intersection / (intersection + false_negatives)
            if false_negatives != 0
            else 1
        )

        # If the union is 0, neither ground_truth nor prediction contain anything -> IoU becomes 1
        iou = intersection / union if union != 0 else 1

        # compute absolute area difference
        # abs_area_diff = np.abs(np.sum(drusen_gt) - np.sum(drusen_pred))
        # Normalize by the number of AScans with drusen
        # drusen_ascans = np.sum(np.sum(union, axis=0) > 0) if np.any(union) else 1.0
        # adad = abs_area_diff / drusen_ascans * 100
        # result[("ADAD", bscan_gt.name)] = [adad]

        result[("IoU", bscan_gt.name)] = [iou]
        result[("Dice", bscan_gt.name)] = [dice]
        result[("Precision", bscan_gt.name)] = [precision]
        result[("Recall", bscan_gt.name)] = [recall]
        result[("GT Drusen Volume", bscan_gt.name)] = [np.sum(drusen_gt)]
        result[("PR Drusen Volume", bscan_gt.name)] = [np.sum(drusen_pred)]

        drusen_load = np.sum(drusen_gt)
        if drusen_load == 0:
            result[("Drusen Load", bscan_gt.name)] = ["No Drusen"]
        elif drusen_load <= 150:
            result[("Drusen Load", bscan_gt.name)] = ["Drusen<150"]
        elif drusen_load <= 1000:
            result[("Drusen Load", bscan_gt.name)] = ["Drusen<1000"]
        else:
            result[("Drusen Load", bscan_gt.name)] = ["Drusen>1000"]

    return result


def compare_annotations(data_gt, data_pred, layers=["BM", "RPE"]):
    """Compare layer and drusen annotations for two OCT volumes.

    Return two pandas DataFrames. One for BScan wise Area measures and
    one for an AScan wise layer height comparison.
    """

    layer_results = compare_layers_per_ascan(data_gt, data_pred, layers=["BM", "RPE"])
    area_results = compare_drusen_per_bscan(data_gt, data_pred)

    area_df = pd.DataFrame.from_dict(area_results).T
    area_df.index.names = ["Metric", "BScan"]
    area_df = pd.melt(area_df.unstack().T, id_vars=["Drusen Load"], ignore_index=False)
    area_df = area_df.reset_index().set_index(["BScan"]).drop("level_0", axis=1)

    layer_df = pd.DataFrame.from_dict(layer_results).T
    layer_df.index.names = ["Metric", "BScan", "AScan"]
    layer_df = pd.melt(
        layer_df.unstack([-1, -2]).T, id_vars=["Drusen Load"], ignore_index=False
    )
    layer_df = (
        layer_df.reset_index().set_index(["BScan", "AScan"]).drop("level_0", axis=1)
    )

    return area_df, layer_df



# -*- coding: utf-8 -*-
"""Top-level package for eyepy."""

__author__ = """Olivier Morelle"""
__email__ = "oli4morelle@gmail.com"
__version__ = "0.3.7"

from eyepy.core.base import Annotation, Bscan, EnfaceImage, Meta, Oct



# -*- coding: utf-8 -*-
from .v103 import bscan_spec as v103_bscan
from .v103 import oct_spec as v103_oct


def HEVOL_VERSIONS(version):
    versions = {"HSF-OCT-103": v103_oct}
    return versions[version]()


def HEVOL_BSCAN_VERSIONS(version):
    versions = {"HSF-BS-103": v103_bscan}
    return versions[version]()



# -*- coding: utf-8 -*-
import logging
from datetime import datetime, timedelta, timezone
from typing import Union

logger = logging.getLogger(__name__)


def _get_meta_attr(meta_attr):
    def prop_func(self):
        return getattr(self.meta, meta_attr)

    return property(prop_func)


def _clean_ascii(unpacked: tuple):
    return unpacked[0].decode("ascii").rstrip("\x00")


def _get_first(unpacked: tuple):
    return unpacked[0]


def _date_in_seconds(
    dt: datetime,
    epoche: datetime = datetime.utcfromtimestamp(0),
    second_frac: Union[float, int] = 1,
):
    seconds = (dt - epoche).total_seconds() / second_frac
    if not seconds.is_integer():
        raise ValueError("The resulting number needs to be a whole number")
    return int(seconds)


def _date_from_seconds(
    s: int,
    epoche: datetime = datetime.utcfromtimestamp(0),
    second_frac: Union[float, int] = 1,
):
    return epoche + timedelta(seconds=s * second_frac)


def _get_first_as_int(elements):
    if elements:
        return int(elements[0].text)
    else:
        return None


def _get_first_as_float(elements):
    if elements:
        return float(elements[0].text)
    else:
        return None


def _get_first_as_str(elements):
    if elements:
        return elements[0].text
    else:
        return None


def _get_datetime_from_xml(elements):
    date = elements[0].find("Series/ExamDate/Date")
    time = elements[0].find("Image/AcquisitionTime/Time")
    if date is None or time is None:
        return None

    year = int(date.find("Year").text)
    month = int(date.find("Month").text)
    day = int(date.find("Day").text)
    hour = int(time.find("Hour").text)
    minute = int(time.find("Minute").text)
    second = int(float(time.find("Second").text))
    utc_bias = int(time.find("UTCBias").text)
    tz = timezone(timedelta(minutes=utc_bias))
    return datetime(year, month, day, hour, minute, second, tzinfo=tz)


def _get_date_from_xml(elements):
    date = elements[0]
    year = int(date.find("Year").text)
    month = int(date.find("Month").text)
    day = int(date.find("Day").text)
    return datetime(year, month, day).date()



# -*- coding: utf-8 -*-
import logging
from functools import lru_cache
from typing import Tuple, Union

import numpy as np

Shape = Union[int, Tuple[int, int]]

logger = logging.getLogger(__name__)


@lru_cache(maxsize=8)
def filtergrid(
    size: Shape, quadrant_shift: bool = True, normalize: bool = True
) -> Tuple[np.ndarray, np.ndarray]:
    """Generates grid for constructing frequency domain filters.
    Parameters
    ----------
    size : Size of the filter
    quadrant_shift : Quadrant shift such that 0 values / frequencies are at the corners
    normalize: Normalize the range to [-0.5,0.5]
    Returns
    -------
    Coordinate matrices for x and y value for a 2D array. The out can be quadrant shifted
    and / or normalized. This is basically a wrapper around np.meshgrid.

    Inspired by filtergrid.m found at https://www.peterkovesi.com/matlabfns/
    """
    if type(size) is int:
        rows = cols = size
    else:
        rows = size[0]
        cols = size[1]

    range_1 = np.linspace(-(cols // 2), np.floor((cols - 1) / 2), cols)
    range_2 = np.linspace(-(rows // 2), np.floor((rows - 1) / 2), rows)

    if normalize:
        range_1 = range_1 / cols
        range_2 = range_2 / rows

    x, y = np.meshgrid(range_1, range_2)

    # Quadrant shift so that filters are constructed with 0 frequency at the corners
    if quadrant_shift:
        x = np.fft.ifftshift(x)
        y = np.fft.ifftshift(y)

    return x.T, y.T


@lru_cache(maxsize=8)
def radius_filtergrid(
    size: Shape, quadrant_shift: bool = True, normalize: bool = True
) -> np.ndarray:
    """
    Parameters
    ----------
    size : Size of the filter
    quadrant_shift : Quadrant shift such that 0 values / frequencies are at the corners
    normalize: Normalize radius to [0 ,0.5]
    Returns
    -------
    A matrix containing the radius from the center. This radius is in range [0, 0.5] if normalized.
    The result can be quadrant shifted such that the 0 values are in the corners.
    """
    x, y = filtergrid(size, quadrant_shift, normalize)
    radius = np.sqrt(x ** 2 + y ** 2)
    return radius


@lru_cache(maxsize=8)
def theta_filtergrid(size: Shape, quadrant_shift: bool = True) -> np.ndarray:
    """
    Parameters
    ----------
    size : Size of the filter
    quadrant_shift : Quadrant shift such that 0 values / frequencies are at the corners
    Returns
    -------
    A matrix containing the polar angle in radian at the respective position for a circle centered in the matrix.
    The result can be returned quadrant shifted. The angle is 0 for all points on the positive x-axis.
    The angles are pi/2 (90°) and -pi/2 (-90°) on the positive and negative y-axis respectively. On the negative
    x-axis the angle is pi (180°). If you need the angle to be in range [0, 2pi] instead of [-pi, pi], you can simply
    add 2pi whenever the angle is negative.
    """

    y, x = filtergrid(size, quadrant_shift)

    # Matrix values contain polar angle.
    # 0 angle starts on the horizontal line and runs counter clock-wise
    theta = np.arctan2(-y, x)

    return theta



# -*- coding: utf-8 -*-
import hashlib
import logging
import os
import warnings
from collections.abc import MutableMapping
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import imageio
import numpy as np
from matplotlib import cm, colors, patches
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from skimage import transform

from eyepy.core import config
from eyepy.core.drusen import DefaultDrusenFinder, DrusenFinder
from eyepy.core.quantifier import DefaultEyeQuantifier, EyeQuantifier

logger = logging.getLogger(__name__)


class Meta(MutableMapping):
    def __init__(self, *args, **kwargs):
        """The Meta object is a dict with additional functionalities.

        The additional functionallities are:
        1. A string representation suitable for printing the meta information.

        2. Checking if a keys value is a callable before returning it. In case
        it is a callable, it sets the value to the return value of the callable.
        This is used for lazy loading OCT data. The meta information for the OCT
        and all B-Scans is only read from the file when accessed.

        An instance of the meta object can be created as you would create an
        ordinary dictionary.

        For example:

            + Meta({"SizeX": 512})
            + Meta(SizeX=512)
            + Meta([(SizeX, 512), (SizeY, 512)])
        """
        self._store = dict()
        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key):
        value = self._store[key]
        if callable(value):
            self[key] = value()
        return self._store[key]

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __str__(self):
        return f"{os.linesep}".join([f"{f}: {self[f]}" for f in self if f != "__empty"])

    def __repr__(self):
        return self.__str__()


class EnfaceImage:
    def __init__(self, data, name=None):
        self._data = data
        self._name = name

    @property
    def data(self):
        """Return the enface image as numpy array."""
        if callable(self._data):
            self._data = self._data()
        return self._data

    @property
    def name(self):
        if self._name is None:
            raise ValueError("This EnfaceImage has no respective filename")
        else:
            return self._name


class Annotation(MutableMapping):
    def __init__(self, *args, **kwargs):
        self._store = dict()
        self.update(dict(*args, **kwargs))  # use the free update to set keys
        self._bscan = None

    def __getitem__(self, key):
        value = self._store[key]
        if callable(value):
            self[key] = value(self.bscan)
        return self._store[key]

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    # TODO: Make the annotation printable to get an overview
    # def __str__(self):
    #     return f"{os.linesep}".join(
    #         [f"{f}: {self[f]}" for f in self if f != "__empty"])

    # def __repr__(self):
    #     return self.__str__()

    @property
    def bscan(self):
        if self._bscan is None:
            raise AttributeError("bscan is not set for this Annotation.")
        return self._bscan

    @bscan.setter
    def bscan(self, value: "Bscan"):
        self._bscan = value


class LayerAnnotation(MutableMapping):
    def __init__(self, data, layername_mapping=None, max_height=2000):
        self._data = data
        self.max_height = max_height
        if layername_mapping is None:
            self.mapping = config.SEG_MAPPING
        else:
            self.mapping = layername_mapping

    @property
    def data(self):
        if callable(self._data):
            self._data = self._data()
        return self._data

    def __getitem__(self, key):
        data = self.data[self.mapping[key]]
        nans = np.isnan(data)
        empty = np.nonzero(
            np.logical_or(
                np.less(data, 0, where=~nans),
                np.greater(data, self.max_height, where=~nans),
            )
        )
        data = np.copy(data)
        data[empty] = np.nan
        if np.nansum(data) > 0:
            return data
        else:
            raise KeyError(f"There is no data given for the {key} layer")

    def __setitem__(self, key, value):
        self.data[self.mapping[key]] = value

    def __delitem__(self, key):
        self.data[self.mapping[key], :] = np.nan

    def __iter__(self):
        inv_map = {v: k for k, v in self.mapping.items()}
        return iter(inv_map.values())

    def __len__(self):
        return len(self.data.shape[0])

    def layer_indices(self, key):
        layer = self[key]
        nan_indices = np.isnan(layer)
        col_indices = np.arange(len(layer))[~nan_indices]
        row_indices = np.rint(layer).astype(int)[~nan_indices]

        return (row_indices, col_indices)


class Bscan:
    def __new__(
        cls,
        data,
        annotation=None,
        meta=None,
        data_processing=None,
        oct_obj=None,
        name=None,
        *args,
        **kwargs,
    ):
        # Make all meta fields accessible as attributes of the BScan without
        # reading them. Set a property instead
        def meta_func_builder(x):
            return lambda self: self.meta[x]

        if meta is not None:
            for key in meta:
                setattr(cls, key, property(meta_func_builder(key)))
        return object.__new__(cls, *args, **kwargs)

    def __init__(
        self,
        data: Union[np.ndarray, Callable],
        annotation: Optional[Annotation] = None,
        meta: Optional[Union[Dict, Meta]] = None,
        data_processing: Optional[Callable] = None,
        oct_obj: Optional["Oct"] = None,
        name: Optional[str] = None,
    ):
        """

        Parameters
        ----------
        data : A numpy array holding the raw B-Scan data or a callable which
            returns a raw B-Scan. Raw means that it represents the unprocessed
            stored data. The actual dtype and value range depends on the storage
            format.
        annotation: Dict holding B-Scan annotations
        meta : A dictionary holding the B-Scans meta informations or
        oct_obj : Reference to the OCT Volume holding the B-Scan
        name : Filename of the B-Scan if B-Scan is save as individual file
        """
        self._scan_raw = data
        self._scan = None
        self._meta = meta
        self._oct_obj = oct_obj

        self._annotation = annotation

        if data_processing is None:
            self._data_processing = lambda x: x
        else:
            self._data_processing = data_processing

        self._name = name

    @property
    def oct_obj(self):
        if self._oct_obj is None:
            raise AttributeError("oct_obj is not set for this Bscan object")
        return self._oct_obj

    @oct_obj.setter
    def oct_obj(self, value):
        self._oct_obj = value

    @property
    def name(self):
        if self._name is None:
            self._name = str(self.index)
        return self._name

    @property
    def index(self):
        return self.oct_obj.bscans.index(self)

    @property
    def meta(self):
        """A dict holding all Bscan meta data."""
        return self._meta

    @property
    def annotation(self):
        """A dict holding all Bscan annotation data."""
        if self._annotation is None:
            self._annotation = Annotation({})
        elif callable(self._annotation):
            self._annotation = self._annotation()
        self._annotation.bscan = self
        return self._annotation

    @property
    def scan_raw(self):
        """An array holding a single raw B-Scan.

        The dtype is not changed after the import. If available this is
        the unprocessed output of the OCT device. In any case this is
        the unprocessed data imported by eyepy.
        """
        if callable(self._scan_raw):
            self._scan_raw = self._scan_raw()
        return self._scan_raw

    @property
    def scan(self):
        """An array holding a single B-Scan with the commonly used contrast.

        The array is of dtype <ubyte> and encodes the intensities as
        values between 0 and 255.
        """
        if self._scan is None:
            self._scan = self._data_processing(self.scan_raw)
        return self._scan

    @property
    def shape(self):
        return self.scan.shape

    @property
    def layers(self):
        if "layers" not in self.annotation:
            l_shape = np.zeros(
                (max(config.SEG_MAPPING.values()) + 1, self.oct_obj.SizeX)
            )
            self.annotation["layers"] = LayerAnnotation(l_shape)
        if callable(self.annotation["layers"]):
            self.annotation["layers"] = self.annotation["layers"]()
        return self.annotation["layers"]

    @property
    def drusen_raw(self):
        """Return drusen computed from the RPE and BM layer segmentation.

        The raw drusen are computed based on single B-Scans
        """
        return self._oct_obj.drusen_raw[..., self.index]

    @property
    def drusen(self):
        """Return filtered drusen.

        Drusen are filtered based on the complete volume
        """
        return self._oct_obj.drusen[..., self.index]

    def plot(
        self,
        ax=None,
        layers=None,
        drusen=False,
        layers_kwargs=None,
        layers_color=None,
        annotation_only=False,
        region=np.s_[:, :],
    ):
        """Plot B-Scan with segmented Layers."""
        if ax is None:
            ax = plt.gca()

        # Complete region index expression
        if region[0].start is None:
            r0_start = 0
        else:
            r0_start = region[0].start
        if region[1].start is None:
            r1_start = 0
        else:
            r1_start = region[1].start
        if region[0].stop is None:
            r0_stop = self.shape[0]
        else:
            r0_stop = region[0].stop
        if region[1].stop is None:
            r1_stop = self.shape[1]
        else:
            r1_stop = region[1].stop
        region = np.s_[r0_start:r0_stop, r1_start:r1_stop]

        if layers is None:
            layers = []
        elif layers == "all":
            layers = self.layers.keys()

        if layers_kwargs is None:
            layers_kwargs = config.layers_kwargs
        else:
            layers_kwargs = {**config.layers_kwargs, **layers_kwargs}

        if layers_color is None:
            layers_color = config.layers_color
        else:
            layers_color = {**config.layers_color, **layers_color}

        if not annotation_only:
            ax.imshow(self.scan[region], cmap="gray")
        if drusen:
            visible = np.zeros(self.drusen.shape)
            visible[self.drusen] = 1.0
            ax.imshow(self.drusen[region], alpha=visible[region], cmap="Reds")
        for layer in layers:
            color = layers_color[layer]
            try:
                layer_data = self.layers[layer]
                # Adjust layer height to plotted region
                layer_data = layer_data - region[0].start
                # Remove layer if outside of region
                layer_data = layer_data[region[1].start : region[1].stop]
                layer_data[layer_data < 0] = 0
                region_height = region[0].stop - region[0].start
                layer_data[layer_data > region_height] = region_height

                ax.plot(
                    layer_data, color=color, label=layer, **layers_kwargs,
                )
            except KeyError:
                warnings.warn(f"Layer '{layer}' has no Segmentation", UserWarning)


class Oct:
    """.vol header
    -----------
    All fields from the .vol header (the oct meta data) can be accessed as attributes of the
    HeyexOct object.

    SLO
    ---
    The attribute `slo` of the HeyexOct object gives access to the IR SLO image
    and returns it as a numpy.ndarray of dtype `uint8`.

    B-Scans
    -------
    Individual B-Scans can be accessed using `oct_scan[index]`. The returned
    HeyexBscan object exposes all B-Scan header fields as attributes and the
    raw B-Scan image as `numpy.ndarray` of type `float32` under the attribute
    `scan_raw`. A transformed version of the raw B-Scan which is more similar to
    the Heyex experience can be accessed with the attribute `scan` and returns
    the 4th root of the raw B-Scan scaled to [0,255] as `uint8`.

    Segmentations
    -------------
    B-Scan segmentations can be accessed for individual B-Scans like
    `bscan.segmentation`. This return a numpy.ndarray of shape (NumSeg, SizeX)
    The `segmentation` attribute of the HeyexOct object returns a dictionary,
    where the key is a number and the value is a numpy.ndarray of shape
    (NumBScans, SizeX)."""

    def __new__(
        cls,
        bscans: List[Bscan],
        localizer=None,
        meta=None,
        data_path=None,
        *args,
        **kwargs,
    ):
        # Set all the meta fields as attributes

        if meta is not None:

            def meta_func_builder(x):
                return lambda self: self.meta[x]

            for key in meta:
                # Every key in meta becomes a property for the new class. The
                # property calls self.meta[key] to retrieve the keys value

                # This lambda func returns a lambda func which accesses the meta
                # object for the key specified in the first lambda. Like this we
                # read the file only on access.
                setattr(cls, key, property(meta_func_builder(key)))

        return object.__new__(cls, *args, **kwargs)

    def __init__(
        self,
        bscans: List[Union[Callable, Bscan]],
        localizer: Optional[Union[Callable, EnfaceImage]] = None,
        meta: Optional[Meta] = None,
        drusenfinder: DrusenFinder = DefaultDrusenFinder(),
        eyequantifier: EyeQuantifier = DefaultEyeQuantifier(),
        data_path: Optional[str] = None,
    ):
        """

        Parameters
        ----------
        bscans :
        meta :
        drusenfinder :
        """
        self.bscans = bscans
        self._localizer = localizer
        self._meta = meta
        self._drusenfinder = drusenfinder
        self._eyequantifier = eyequantifier
        self._tform_localizer_to_oct = None

        self._drusen = None
        self._drusen_raw = None

        self._eyepy_id = None
        if data_path is None:
            self.data_path = Path.home() / ".eyepy"
        self.data_path = Path(data_path)
        self.drusen_path = self.data_path / ".eyepy" / f"{self.eyepy_id}_drusen_map.npy"

    def __getitem__(self, index) -> Bscan:
        """The B-Scan at the given index."""
        if type(index) == slice:
            return [self[i] for i in range(*index.indices(len(self)))]
        else:
            bscan = self.bscans[index]
            if callable(bscan):
                self.bscans[index] = bscan()
            self.bscans[index].oct_obj = self
            return self.bscans[index]

    def __len__(self):
        """The number of B-Scans."""
        return len(self.bscans)

    @classmethod
    def from_heyex_xml(cls, path):
        from eyepy.io.heyex import HeyexXmlReader

        reader = HeyexXmlReader(path)
        return cls(
            bscans=reader.bscans,
            localizer=reader.localizer,
            meta=reader.oct_meta,
            data_path=reader.path,
        )

    @classmethod
    def from_heyex_vol(cls, path):
        from eyepy.io.heyex import HeyexVolReader

        reader = HeyexVolReader(path)
        return cls(
            bscans=reader.bscans,
            localizer=reader.localizer,
            meta=reader.oct_meta,
            data_path=Path(path).parent,
        )

    @classmethod
    def from_duke_mat(cls, path):
        import scipy.io as sio

        loaded = sio.loadmat(path)
        data = np.moveaxis(loaded["images"], -1, 0)
        label = np.swapaxes(loaded["layerMaps"], 1, 2)

        bscans = []
        mapping = {"BM": 2, "RPE": 1, "ILM": 0}
        for d, l in zip(data, label):
            annotation = Annotation({"layers": LayerAnnotation(l, mapping)})
            bscans.append(Bscan(d, annotation=annotation))
        return cls(
            bscans=bscans,
            meta=Meta(**{"Age": loaded["Age"]}),
            data_path=Path(path).parent,
        )

    @classmethod
    def from_folder(cls, path):
        path = Path(path)
        img_paths = sorted(list(path.iterdir()))

        def read_func(p):
            return lambda: imageio.imread(p)

        bscans = [Bscan(read_func(p), name=p.name) for p in img_paths]
        return cls(bscans=bscans, data_path=path)

    def estimate_bscan_distance(self):
        # Try to estimate B-Scan distances. Can be used if Bscan Positions
        # but not their distance is in the meta information

        # Pythagoras in case B-Scans are rotated with respect to the localizer
        a = self[-1].StartY - self[0].StartY
        b = self[-1].StartX - self[0].StartX
        self.meta["Distance"] = np.sqrt(a ** 2 + b ** 2) / (len(self.bscans) - 1)
        return self.Distance

    @property
    def eyepy_id(self):
        """Visit ID for saving visit related files."""
        if self._eyepy_id is None:
            # Compute a hash of the first B-Scan as ID
            sha1 = hashlib.sha1()
            sha1.update(self[0].scan.tobytes())
            self._eyepy_id = sha1.hexdigest()
        return self._eyepy_id

    @property
    def fovea_pos(self):
        return None

    @property
    def shape(self):
        return (self.SizeZ, self.SizeX, self.NumBScans)

    @property
    def SizeX(self):
        try:
            return self.meta["SizeX"]
        except:
            return self[0].scan.shape[1]

    @property
    def SizeZ(self):
        try:
            return self.meta["SizeZ"]
        except:
            return self[0].scan.shape[0]

    @property
    def NumBScans(self):
        try:
            return self.meta["NumBScans"]
        except:
            return len(self)

    @property
    def localizer(self):
        """A numpy array holding the OCTs localizer enface if available."""
        try:

            return self._localizer.data
        except AttributeError:
            raise AttributeError("This OCT object has no localizer image.")

    @property
    def volume_raw(self):
        """An array holding the OCT volume.

        The dtype is not changed after the import. If available this is
        the unprocessed output of the OCT device. In any case this is
        the unprocessed data imported by eyepy.
        """
        return np.stack([x.scan_raw for x in self.bscans], axis=-1)

    @property
    def volume(self):
        """An array holding the OCT volume with the commonly used contrast.

        The array is of dtype <ubyte> and encodes the intensities as
        values between 0 and 255.
        """
        return np.stack([x.scan for x in self.bscans], axis=-1)

    @property
    def layers_raw(self):
        """Height maps for all layers combined into one volume.

        Layers for all B-Scans are stacked such that we get a volume L x B x W
        where L are different Layers, B are the B-Scans and W is the Width of
        the B-Scans.

        A flip on the B-Scan axis is needed to locate the first B-Scan at the
        bottom of the height map.
        """
        return np.flip(np.stack([x.layers.data for x in self], axis=1), axis=1)

    @property
    def layers(self):
        """Height maps for all layers accessible by the layers name."""
        nans = np.isnan(self.layers_raw)
        empty = np.nonzero(
            np.logical_or(
                np.less(self.layers_raw, 0, where=~nans),
                np.greater(self.layers_raw, self.SizeZ, where=~nans),
            )
        )

        data = self.layers_raw.copy()
        data[empty] = np.nan
        return {
            name: data[i, ...]
            for name, i in self[0].layers.mapping.items()
            if np.nansum(data[i, ...]) != 0
        }

    @property
    def meta(self):
        """A dict holding all OCT meta data.

        The object can be printed to see all available meta data.
        """
        if self._meta is None:
            raise AttributeError("This volume has no meta data")
        return self._meta

    @property
    def drusen(self):
        """Final drusen after post processing the initial raw drusen.

        Here the `filter` function of the DrusenFinder has been applied
        """
        if self._drusen is None:
            # Try to load the drusen from the default location
            try:
                self._drusen = np.load(self.drusen_path)
            except (NotADirectoryError, FileNotFoundError):
                self._drusen = self._drusenfinder.filter(self.drusen_raw)
                if config.SAVE_DRUSEN:
                    self.drusen_path.parent.mkdir(parents=True, exist_ok=True)
                    np.save(self.drusen_path, self._drusen)
        return self._drusen

    def drusen_recompute(self, drusenfinder=None):
        """Recompute Drusen optionally with a custom DrusenFinder.

        Use this if you do not like the computed / loaded drusen
        """
        if drusenfinder is not None:
            self._drusenfinder = drusenfinder

        self._drusen_raw = self._drusenfinder.find(self)
        self._drusen = self._drusenfinder.filter(self.drusen_raw)
        if config.SAVE_DRUSEN:
            np.save(self.drusen_path, self._drusen)
        return self._drusen

    @property
    def drusen_raw(self):
        """Initial drusen before post procssing.

        The initial drusen found by the DrusenFinders `find` function.
        """
        if self._drusen_raw is None:
            self._drusen_raw = self._drusenfinder.find(self)
        return self._drusen_raw

    @property
    def quantification(self):
        return self._eyequantifier.quantify(self)

    @property
    def tform_localizer_to_oct(self):
        if self._tform_localizer_to_oct is None:
            self._tform_localizer_to_oct = self._estimate_localizer_to_oct_tform()
        return self._tform_localizer_to_oct

    @property
    def tform_oct_to_localizer(self):
        return self.tform_localizer_to_oct.inverse

    @property
    def localizer_shape(self):
        try:
            return self.localizer.shape
        except:
            return (self.SizeX, self.SizeX)

    def _estimate_localizer_to_oct_tform(self):
        oct_projection_shape = (self.NumBScans, self.SizeX)
        src = np.array(
            [
                oct_projection_shape[0] - 1,
                0,  # Top left
                oct_projection_shape[0] - 1,
                oct_projection_shape[1] - 1,  # Top right
                0,
                0,  # Bottom left
                0,
                oct_projection_shape[1] - 1,  # Bottom right
            ]
        ).reshape((-1, 2))
        src = np.array(
            [
                0,
                0,  # Top left
                0,
                oct_projection_shape[1] - 1,  # Top right
                oct_projection_shape[0] - 1,
                0,  # Bottom left
                oct_projection_shape[0] - 1,
                oct_projection_shape[1] - 1,  # Bottom right
            ]
        ).reshape((-1, 2))

        try:
            # Try to map the oct projection to the localizer image
            dst = np.array(
                [
                    self[-1].StartY / self.ScaleXSlo,
                    self[-1].StartX / self.ScaleYSlo,
                    self[-1].EndY / self.ScaleXSlo,
                    self[-1].EndX / self.ScaleYSlo,
                    self[0].StartY / self.ScaleXSlo,
                    self[0].StartX / self.ScaleYSlo,
                    self[0].EndY / self.ScaleXSlo,
                    self[0].EndX / self.ScaleYSlo,
                ]
            ).reshape((-1, 2))
        except AttributeError:
            # Map the oct projection to a square area of shape (bscan_width, bscan_width)
            warnings.warn(
                f"Bscan positions on localizer image or the scale of the "
                f"localizer image is missing. We assume that the B-Scans cover "
                f"a square area and are equally spaced.",
                UserWarning,
            )
            b_width = self[0].shape[1]
            dst = np.array(
                [
                    0,
                    0,  # Top left
                    0,
                    b_width - 1,  # Top right
                    b_width - 1,
                    0,  # Bottom left
                    b_width - 1,
                    b_width - 1,  # Bottom right
                ]
            ).reshape((-1, 2))

        src = src[:, [1, 0]]
        dst = dst[:, [1, 0]]
        tform = transform.estimate_transform("affine", src, dst)

        if not np.allclose(tform.inverse(tform(src)), src):
            msg = f"Problem with transformation of OCT Projection to the localizer image space."
            raise ValueError(msg)

        return tform

    @property
    def drusen_projection(self):
        # Sum the all B-Scans along their first axis (B-Scan height)
        # Swap axis such that the volume depth becomes the projections height not width
        # We want the first B-Scan to be located at the bottom hence flip along axis 0
        return np.flip(np.swapaxes(np.sum(self.drusen, axis=0), 0, 1), axis=0)

    @property
    def drusen_enface(self):
        """Drusen projection warped into the localizer space."""
        return transform.warp(
            self.drusen_projection.astype(float),
            self.tform_oct_to_localizer,
            output_shape=self.localizer_shape,
            order=0,
        )

    @property
    def drusenfinder(self):
        """Get and set the DrusenFinder object.

        When the DrusenFinder object is set all drusen are removed.
        """
        return self._drusenfinder

    @drusenfinder.setter
    def drusenfinder(self, drusenfinder):
        self._drusen = None
        self._drusen_raw = None
        self._drusenfinder = drusenfinder

    def plot(
        self,
        ax=None,
        localizer=True,
        drusen=False,
        bscan_region=False,
        bscan_positions=None,
        masks=False,
        region=np.s_[...],
        drusen_kwargs=None,
    ):
        """

        Parameters
        ----------
        ax :
        slo :
        drusen :
        bscan_region :
        bscan_positions :
        masks :
        region : slice object
        alpha :

        Returns
        -------

        """

        if ax is None:
            ax = plt.gca()

        if localizer:
            self.plot_localizer(ax=ax, region=region)
        if drusen:
            if drusen_kwargs is None:
                drusen_kwargs = {}
            self.plot_drusen(ax=ax, region=region, **drusen_kwargs)
        if bscan_positions is not None:
            self.plot_bscan_positions(
                ax=ax,
                bscan_positions=bscan_positions,
                region=region,
                line_kwargs={"linewidth": 0.5, "color": "green"},
            )
        if bscan_region:
            self.plot_bscan_region(region=region, ax=ax)

        if masks:
            self.plot_masks(region=region, ax=ax)
        # if quantification:
        #    self.plot_quantification(space=space, region=region, ax=ax,
        #    q_kwargs=q_kwargs)

    def plot_bscan_ticks(self, ax=None):
        if ax is None:
            ax = plt.gca()
        ax.yticks()

    def plot_layer_distance(
        self,
        region=np.s_[...],
        ax=None,
        bot_layer="BM",
        top_layer="RPE",
        vmin=None,
        vmax=None,
    ):
        if ax is None:
            ax = plt.gca()

        dist = self.layers["BM"] - self.layers["RPE"]
        img = transform.warp(
            dist.astype(float),
            self.tform_oct_to_localizer,
            output_shape=self.localizer_shape,
            order=0,
        )
        ax.imshow(img[region], cmap="gray", vmin=vmin, vmax=vmax)

    def plot_masks(self, region=np.s_[...], ax=None, color="r", linewidth=0.5):
        """

        Parameters
        ----------
        region :
        ax :
        color :
        linewidth :

        Returns
        -------

        """
        primitives = self._eyequantifier.plot_primitives(self)
        if ax is None:
            ax = plt.gca()

        for circle in primitives["circles"]:
            c = patches.Circle(
                circle["center"],
                circle["radius"],
                facecolor="none",
                edgecolor=color,
                linewidth=linewidth,
            )
            ax.add_patch(c)

        for line in primitives["lines"]:
            x = [line["start"][0], line["end"][0]]
            y = [line["start"][1], line["end"][1]]
            ax.plot(x, y, color=color, linewidth=linewidth)

    def plot_localizer(self, ax=None, region=np.s_[...]):
        if ax is None:
            ax = plt.gca()
        ax.imshow(self.localizer[region], cmap="gray")

    def plot_bscan_positions(
        self, bscan_positions="all", ax=None, region=np.s_[...], line_kwargs=None
    ):
        if bscan_positions is None:
            bscan_positions = []
        elif bscan_positions == "all" or bscan_positions is True:
            bscan_positions = range(0, len(self))

        if line_kwargs is None:
            line_kwargs = config.line_kwargs
        else:
            line_kwargs = {**config.line_kwargs, **line_kwargs}

        for i in bscan_positions:
            bscan = self[i]
            x = np.array([bscan.StartX, bscan.EndX]) / self.ScaleXSlo
            y = np.array([bscan.StartY, bscan.EndY]) / self.ScaleYSlo

            ax.plot(x, y, **line_kwargs)

    def plot_bscan_region(self, region=np.s_[...], ax=None):

        if ax is None:
            ax = plt.gca()

        up_right_corner = (
            self[-1].EndX / self.ScaleXSlo,
            self[-1].EndY / self.ScaleYSlo,
        )
        width = (self[0].StartX - self[0].EndX) / self.ScaleXSlo
        height = (self[0].StartY - self[-1].EndY) / self.ScaleYSlo
        # Create a Rectangle patch
        rect = patches.Rectangle(
            up_right_corner, width, height, linewidth=1, edgecolor="r", facecolor="none"
        )

        # Add the patch to the Axes
        ax.add_patch(rect)

    def plot_drusen(
        self,
        ax=None,
        region=np.s_[...],
        cmap="Reds",
        vmin=None,
        vmax=None,
        cbar=True,
        alpha=1,
    ):
        drusen = self.drusen_enface

        if ax is None:
            ax = plt.gca()

        if vmin is None:
            vmin = 1
        if vmax is None:
            vmax = max([drusen.max(), vmin])

        visible = np.zeros(drusen[region].shape)
        visible[np.logical_and(vmin <= drusen[region], drusen[region] <= vmax)] = 1

        if cbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            plt.colorbar(
                cm.ScalarMappable(colors.Normalize(vmin=vmin, vmax=vmax), cmap=cmap),
                cax=cax,
            )

        ax.imshow(
            drusen[region],
            alpha=visible[region] * alpha,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )

    def plot_localizer_bscan(self, ax=None, n_bscan=0):
        """Plot Slo with one selected B-Scan."""
        raise NotImplementedError()

    def plot_bscans(
        self, bs_range=range(0, 8), cols=4, layers=None, layers_kwargs=None
    ):
        """Plot a grid with B-Scans."""
        rows = int(np.ceil(len(bs_range) / cols))
        if layers is None:
            layers = []

        fig, axes = plt.subplots(cols, rows, figsize=(rows * 4, cols * 4))

        with np.errstate(invalid="ignore"):
            for i in bs_range:
                bscan = self[i]
                ax = axes.flatten()[i]
                bscan.plot(ax=ax, layers=layers, layers_kwargs=layers_kwargs)



# -*- coding: utf-8 -*-
import logging
from typing import List
from abc import ABC, abstractmethod

from skimage import transform

from eyepy.utils.masks import create_region_masks, create_region_shape_primitives

logger = logging.getLogger(__name__)


class EyeQuantifier(ABC):
    @abstractmethod
    def regions(self, oct_obj):
        raise NotImplementedError()

    @abstractmethod
    def quantify(self, oct_obj):
        raise NotImplementedError()

    @abstractmethod
    def plot_primitives(self, oct_obj):
        raise NotImplementedError()


class DefaultEyeQuantifier(EyeQuantifier):
    def __init__(self):
        self._regions = None
        self.radii = [0.8, 1.8]

    def regions(self, oct_obj):
        """A dict `region_name:mask` for regions in the OCT."""
        if self._regions is None:
            radii_enface_space = [r / oct_obj.ScaleXSlo for r in self.radii]
            masks = create_region_masks(
                (oct_obj.SizeXSlo, oct_obj.SizeYSlo),
                radii=radii_enface_space,
                n_sectors=[4, 1, 1],
                rotation=[45, 0, 0],
                center=oct_obj.fovea_pos,
                smooth_edges=False,
                ring_sectors=True,
                add_circle_masks=True,
            )

            region_names = {
                "OS": ["Superior", "Nasal", "Inferior", "Temporal"],
                "OD": ["Superior", "Temporal", "Inferior", "Nasal"],
            }
            names = region_names[oct_obj.ScanPosition]
            names = [f"{n} {self.radii[0]}-{self.radii[1]}" for n in names]
            [names.append(f"Radius-{r}") for r in self.radii]
            self._regions = {name: mask for name, mask in zip(names, masks)}
        return self._regions

    @property
    def radii(self):
        return self._radii

    @radii.setter
    def radii(self, value: List):
        self._radii = value
        self._regions = None

    def plot_primitives(self, oct_obj):
        radii_enface_space = [r / oct_obj.ScaleXSlo for r in self.radii]
        return create_region_shape_primitives(
            (oct_obj.SizeXSlo, oct_obj.SizeYSlo),
            radii=radii_enface_space,
            n_sectors=[4, 1, 1],
            rotation=[45, 0, 0],
            center=oct_obj.fovea_pos,
        )

    def quantify(self, oct_obj, space="enface"):
        """Quantify the defined regions.

        B-Scan Voxels are often stretched perpendicular to the B-Scan direction.

        Parameters
        ----------
        oct_obj :
        space : Either `enface` or `oct`. This specifies whether to quantify in
        the enface space by warping the oct drusen projection on the enface or
        in oct space, by warping the quantification masks on the oct drusen
        projection. `enface` is a little faster since only one image and not all
        masks have to be warped.

        Returns
        -------
        """
        oct_voxel_size_µm3 = (
            oct_obj.ScaleX * 1e3 * oct_obj.Distance * 1e3 * oct_obj.ScaleZ * 1e3
        )
        enface_voxel_size_µm3 = (
            oct_obj.ScaleXSlo * 1e3 * oct_obj.ScaleYSlo * 1e3 * oct_obj.ScaleZ * 1e3
        )

        if space == "enface":
            drusen_enface = oct_obj.drusen_enface
            voxel_size_µm3 = enface_voxel_size_µm3
        elif space == "oct":
            drusen_enface = oct_obj.drusen_projection
            voxel_size_µm3 = oct_voxel_size_µm3
        else:
            raise ValueError("Choose 'oct' or 'enface' for the space parameter")

        masks = self.regions(oct_obj)
        results = {}
        for name, mask in masks.items():
            # The voxel size in enface space not in the measurement (oct) space
            if space == "oct":
                mask = transform.warp(mask, oct_obj.tform_localizer_to_oct)[
                    : oct_obj.NumBScans, : oct_obj.SizeX
                ]
            results[f"{name} [mm³]"] = (
                (drusen_enface * mask).sum() * voxel_size_µm3 / 1e9
            )

        results["Total [mm³]"] = drusen_enface.sum() * voxel_size_µm3 / 1e9
        results["Total [OCT voxels]"] = oct_obj.drusen_projection.sum()
        results["OCT Voxel Size [µm³]"] = oct_voxel_size_µm3
        # results["Enface Voxel Size [µm³]"] = enface_voxel_size_µm3
        return results


# -*- coding: utf-8 -*-
import logging
from abc import ABC, abstractmethod

import numpy as np

import scipy.ndimage as ndimage
from scipy.interpolate import interp1d

logger = logging.getLogger(__name__)


class DrusenFinder(ABC):
    @abstractmethod
    def find(self, oct_obj):
        """A function which returns a boolean map indicating drusen."""
        raise NotImplementedError()

    @abstractmethod
    def filter(self, drusen_map):
        """A function which filters drusen based on the drusen properties."""
        raise NotImplementedError()

class DefaultDrusenFinder(DrusenFinder):
    def __init__(
        self,
        minimum_height=2,
        voxel_size=(1, 1, 1),
    ):
        """

        Parameters
        ----------
        minimum_height: Minimum drusen  height.
        voxel_size: (height, width, depth) length of the voxel in µm.
        """
        self.minimum_height = minimum_height
        self.voxel_size = voxel_size

    def filter(self, drusen_map):
        return filter_by_height(drusen_map, self.minimum_height,
                                    self.voxel_size)

    def find(self, oct_obj):
        return drusen2d(oct_obj.layers["RPE"], oct_obj.layers["BM"],
                        oct_obj.shape)

def drusen2d(
    rpe_height,
    bm_height,
    volume_shape):
    """Compute drusen from the RPE and BM layer segmentation.

    First estimate the ideal RPE based on a histogram of the RPE heights relativ
    to the BM.
    Then compute drusen as the area between the RPE and the normal RPE
    """
    # Estimate ideal RPE
    ideal_rpe = ideal_rpe2d(rpe_height, bm_height, volume_shape)

    # Create drusen map
    drusen_map = np.zeros(volume_shape, dtype=bool)
    # Exclude normal RPE and RPE from the drusen area.
    rpe = np.flip((rpe_height + 1).astype(int), axis=0)
    irpe = np.flip(ideal_rpe.astype(int), axis=0)
    for sli in range(drusen_map.shape[2]):
        for col in range(drusen_map.shape[1]):
            if not rpe[sli, col] == -9223372036854775808:
                drusen_map[rpe[sli, col]: irpe[sli, col], col, sli] = 1

    return drusen_map

def ideal_rpe2d(rpe_height, bm_height, volume_shape):
    h, w, d = volume_shape

    # compute shift needed to align the BM to the horizontal center line
    shift = np.empty((d, w), dtype='int')
    shift.fill(h - (h / 2))
    shift = shift - bm_height

    # now shift the RPE location array as well
    shifted_rpe_height = (rpe_height + shift)

    # Remove all NANs from the shifted RPE data
    clean_shifted = shifted_rpe_height[~np.isnan(shifted_rpe_height)]

    # Compute a histogram with a bin for ever pixel height in a B-Scan
    hist, edges = np.histogram(clean_shifted.flatten(),
                               bins=np.arange(volume_shape[0]))

    # Compute the ideal RPE as the mean of the biggest bin and its neighbours
    lower_edge = edges[np.argmax(hist) - 1]
    upper_edge = edges[np.argmax(hist) + 2]
    irpe_height = np.mean(clean_shifted[
                              np.logical_and(clean_shifted <= upper_edge,
                                             clean_shifted >= lower_edge)])
    ideal_rpe = np.full_like(shifted_rpe_height, irpe_height)

    # Shift back into original image space
    ideal_rpe = np.reshape(ideal_rpe, (d, w)) - shift

    return ideal_rpe


class DrusenFinderPolyFit(DrusenFinder):
    def __init__(
        self,
        degree=3,
        iterations=5,
        outlier_threshold=5,
        poly_fit_type="regularized",
        minimum_height=2,
        voxel_size=(1, 1, 1),
    ):
        """

        Parameters
        ----------
        degree: Maximum degree of the Polynomial used for finding the normal RPE
        iterations: Number of iterations in the outlier removal
        outlier_threshold: Threshold for RPE Pixel to be considered in the next iteration
        poly_fit_type: If "regularized", regularize the normal RPE fitting.
        minimum_height: Minimum drusen  height.
        voxel_size: (height, width, depth) length of the voxel in µm.
        """
        self.degree = degree
        self.iterations = iterations
        self.outlier_threshold = outlier_threshold
        self.poly_fit_type = poly_fit_type
        self.minimum_height = minimum_height
        self.voxel_size = voxel_size

    def filter(self, drusen_map):
        return filter_by_height(drusen_map, self.minimum_height,
                                    self.voxel_size)

    def find(self, oct_obj):
        drusen_map = np.zeros(oct_obj.shape, dtype=bool)
        for i, scan in enumerate(oct_obj):
            try:
                drusen_map[..., i] = drusen(
                    scan.layers["RPE"],
                    scan.layers["BM"],
                    scan.shape,
                    degree=self.degree,
                    iterations=self.iterations,
                    outlier_threshold=self.outlier_threshold,
                    poly_fit_type=self.poly_fit_type,
                )
            except KeyError as e:
                logger.warning(e)

        return drusen_map

def drusen(
    rpe_height,
    bm_height,
    scan_shape,
    degree=3,
    iterations=5,
    outlier_threshold=5,
    poly_fit_type="regularized",
):
    """Compute drusen from the RPE layer segmentation.

    First estimate the ideal RPE by fitting a polynomial to the RPE.
    Then compute drusen as the area between the RPE and the normal RPE
    """

    rpe_height = np.rint(interpolate_layer(rpe_height))
    bm_height = np.rint(interpolate_layer(bm_height))
    # Estimate normal RPE
    normal_rpe_height = normal_rpe(
        rpe_height,
        bm_height,
        scan_shape,
        degree,
        iterations,
        outlier_threshold,
        poly_fit_type,
    )
    # Create drusen map
    drusen_map = np.zeros(scan_shape, dtype=bool)

    # Exclude normal RPE and RPE from the drusen area.
    for col in range(drusen_map.shape[1]):
        drusen_map[
            (rpe_height + 1).astype(int)[col] : normal_rpe_height.astype(int)[col], col
        ] = 1
    return drusen_map

def interpolate_layer(layer_height, kind="nearest"):
    nans = np.isnan(layer_height)
    x = np.arange(layer_height.shape[0])[~nans]
    y = layer_height[~nans]
    f = interp1d(x, y, kind=kind, fill_value="extrapolate")
    interpolated_layer = np.copy(layer_height)
    interpolated_layer[nans] = f(np.argwhere(nans)[:, 0])
    return interpolated_layer


def normal_rpe(
    rpe_height,
    bm_height,
    scan_shape,
    degree=3,
    iterations=5,
    outlier_threshold=5,
    poly_fit_type="regularized",
):
    """Estimate the normal RPE First the shift to make the BM a horizontal line
    is found.

    Then this shift is applied to the RPE. A third degree polynomial is
    fitted to the shifted RPE and the resulting normal RPE curve is
    shifted back into the original image space.
    """
    h, w = scan_shape

    # compute shift needed to align the BM to the horizontal center line
    shift = np.empty(w, dtype="int")
    shift.fill(h - (h / 2))
    shift = shift - bm_height

    # now shift the RPE location vector as well
    shifted_rpe_height = rpe_height + shift

    # These variables change while outliers are removed and polynomial fitting
    tmpx = range(w)
    tmpy = np.copy(shifted_rpe_height)

    it = 0
    while True:
        if poly_fit_type == "regularized":
            coeffs = compute_regularized_fit(tmpx, tmpy, deg=degree)
        else:
            coeffs = np.polyfit(tmpx, tmpy, deg=degree)

        # Evaluate the polynomial for all x values
        norm_rpe = np.polyval(coeffs, range(w)).astype("int")

        # Outlier removal
        if outlier_threshold:
            # Compute norm rpe with these values in next iteration
            inlier = norm_rpe - shifted_rpe_height < outlier_threshold
            tmpx = np.argwhere(inlier)[:, 0]
            tmpy = shifted_rpe_height[inlier]
        else:
            # TODO: Is this correct
            # Element wise maximum of the polynomial fitted rpe and the rpe
            tmpy = np.maximum(norm_rpe, tmpy)

        it += 1
        if it >= iterations:
            break

    # Shift back into original image space
    norm_rpe = norm_rpe - shift

    return norm_rpe

def compute_regularized_fit(x, y, deg):
    result_matrix = np.zeros((deg + 1, deg + 1))
    for d in range(deg + 1):
        z = np.polyfit(x, y, deg=d)
        for i in range(len(z)):
            result_matrix[d, -1 - i] = z[-1 - i]
    # The highest degree has the lowest weight
    weighted_average = np.average(
        result_matrix, axis=0, weights=[1.0, 1.0, 0.1 * 2, 0.1 ** 4]
    )
    return weighted_average

def filter_by_depth(drusen_map, minimum_depth=2):
    filtered_drusen = np.copy(drusen_map)
    if minimum_depth == 0:
        return drusen_map
    # get array where connected components get same label
    connected_component_array, num_drusen = ndimage.label(drusen_map)
    # Go through each component, sum it along 2 axis and check max depth against threshold
    max_depths = np.zeros_like(connected_component_array)
    for label, drusen_pos in enumerate(ndimage.find_objects(connected_component_array)):
        component_sub_vol = connected_component_array[drusen_pos]
        component_max_depth = np.max(np.sum(component_sub_vol == label + 1, axis=2))
        component_sub_vol[component_sub_vol == label + 1] = component_max_depth
        max_depths[drusen_pos] = component_sub_vol
    filtered_drusen[max_depths < minimum_depth] = False
    return filtered_drusen.astype(bool)


def filter_by_height(drusen_map, minimum_height=2, voxel_size=(1, 1, 1)):
    if minimum_height == 0:
        return drusen_map
    connected_component_array, num_drusen = ndimage.label(drusen_map)
    component_height_array = component_max_height(connected_component_array)

    filtered_drusen = np.copy(drusen_map)
    filtered_drusen[component_height_array < minimum_height] = False
    return filtered_drusen.astype(bool)


def component_max_height(connected_component_array):
    max_heights = np.zeros_like(connected_component_array)
    # Iterate over all connected drusen components
    for drusen_pos in ndimage.find_objects(connected_component_array):
        # Work on subvolume for faster processing
        component_sub_vol = connected_component_array[drusen_pos]
        # Find current label (most frequent label in the subvolume)
        label = np.bincount(component_sub_vol[component_sub_vol!=0]).argmax()
        component_max_height = np.max(np.sum(component_sub_vol == label, axis=0))
        # Set drusen region to drusen max height
        max_heights[drusen_pos][component_sub_vol == label] = component_max_height
    return max_heights


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import sys

from setuptools import find_packages, setup

# These are also specified in eyepy.__init__.py
__author__ = """Olivier Morelle"""
__email__ = "oli4morelle@gmail.com"
__version__ = "0.3.7"

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["imageio", "numpy", "matplotlib", "seaborn", "scikit-image",
                "scipy", "imagecodecs"]

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest>=3"]

setup(
    author=__author__,
    author_email=__email__,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="The Python package for working with ophthalmological data.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords=["eyepy", "eyepie"],
    name="eyepie",
    packages=find_packages(include=["eyepy", "eyepy.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/MedVisBonn/eyepy",
    version=__version__,
    zip_safe=False,
)



# -*- coding: utf-8 -*-
import seaborn as sns

# Plotting config

# PR1 and EZ map to 14 and PR2 and IZ map to 15. Hence both names can be used
# to access the same data
SEG_MAPPING = {
    "ILM": 0,
    "BM": 1,
    "RNFL": 2,
    "NFL": 2,
    "GCL": 3,
    "IPL": 4,
    "INL": 5,
    "OPL": 6,
    "ONL": 7,
    "ELM": 8,
    "IOS": 9,
    "OPT": 10,
    "CHO": 11,
    "VIT": 12,
    "ANT": 13,
    "EZ": 14,
    "PR1": 14,
    "IZ": 15,
    "PR2": 15,
    "RPE": 16,
}

SAVE_DRUSEN = False

# Line Style for Layers in B-Scan
layers_kwargs = {"linewidth": 1, "linestyle": "-"}

# Line Style for B-Scan positions on Slo
line_kwargs = {"linewidth": 0.3, "linestyle": "-"}

# Colors for different Layers
x = sns.color_palette("husl", 17)
color_palette = sns.color_palette(x[::3] + x[1::3] + x[2::3])
layers_color = {key: color_palette[value] for key, value in SEG_MAPPING.items()}
